"""Utils to fill templates at the start of the container.

Template must be filled :
    - after volume mounts and environment initialization,
    - before application start script.
"""
import os
from pathlib import Path
from typing import Any

from nua.lib.actions import jinja2_render_file


def render_config_templates(data: dict):
    """Find files and templates in the /nua/templates folder and fill them.

    Destination files may be on Docker-mounted volumes.

    This function is expected to be run as root. Files are stored with
    755 rights. Files ending with .j2 are considered as Jinja2
    templates.
    """
    src_folder = Path("/nua/templates")
    if not src_folder.is_dir():
        return
    data.update(os.environ)
    _iterate_template_files(src_folder, data)


def _iterate_template_files(src_folder: Path, data: dict[str, Any]):
    for file in src_folder.rglob("*"):
        if not file.is_file():
            continue
        if file.name.startswith("."):
            continue
        dest = Path("/") / file.relative_to(src_folder)
        if file.name.endswith(".j2"):
            _render_template(file, dest.with_suffix(""), data)
        else:
            _copy_file(file, dest)


def _copy_file(src: Path, dest: Path):
    dest.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
    dest.write_bytes(src.read_bytes())
    dest.chmod(0o644)


def _render_template(template: Path, dest: Path, data: dict[str, Any]):
    dest.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
    jinja2_render_file(template, dest, data)
    dest.chmod(0o644)



"""Apply automated installation detection heuristics.
"""
from pathlib import Path

from nua.lib.backports import chdir
from nua.lib.panic import info, show
from nua.lib.tool.state import verbosity

from .detectors import PythonSource, PythonWheels

AUTO_INSTALL = [
    PythonSource,
    PythonWheels,
]


def auto_install() -> bool:
    for auto_installer in AUTO_INSTALL:
        if auto_installer.detect():
            with verbosity(2):
                msg = auto_installer.info()
                show(f"{msg} detected")
            auto_installer.install()
            return True
    with verbosity(2):
        show("No automated installation detection")
    return False


def detect_and_install(directory: str | Path | None) -> bool:
    """Apply automated installation detection heuristics."""
    if directory:
        path = Path(directory).resolve()
    else:
        path = Path(".").resolve()
    with verbosity(2):
        info("Detect and install in ", path)
    with chdir(path):
        return auto_install()



"""
Common interface for databases : SQLite3 interface
"""
import sqlite3
from pathlib import Path

from nua.lib.panic import vprint
from nua.lib.tool.state import verbosity

from .db_manager import DbManager


class SQLiteManager(DbManager):
    """Database manager for SQLite3."""

    def __init__(self, host: str, port: str, user: str, password: str):
        """No actual init required for SQLite3."""
        return

    def setup_db_user(self, dbname: str, user: str, password: str):
        return

    def remove_db_user(self, dbname: str, user: str):
        return

    def db_drop(self, dbname: str):
        if dbname == ":memory:":
            return
        Path(dbname).unlink()

    def db_dump(self, dbname: str, **kwargs: str):
        if not self.db_exist(dbname):
            raise ValueError(f"Sqlite database {dbname} doesn't exist.")
        dest_file = kwargs.get("output")
        if not dest_file:
            raise ValueError("Missing 'output' destination file.")
        connection = sqlite3.connect(dbname)
        with open(dest_file, "w") as wfile:
            for line in connection.iterdump():
                wfile.write(f"{line}\n")
        connection.close()

    def user_drop(self, user: str) -> bool:
        return True

    def user_exist(self, user: str) -> bool:
        return True

    def user_create(self, user: str, password: str):
        return

    def db_exist(self, dbname: str) -> bool:
        return Path(dbname).is_file()

    def db_create(self, dbname: str, user: str, _password: str | None = None):
        connection = sqlite3.connect(dbname)
        with verbosity(2):
            vprint(f"SQLite DB {dbname} created.")
        connection.close()

    def db_table_exist(self, dbname: str, user: str, password: str, table: str) -> bool:
        if not self.db_exist(dbname):
            return False
        connection = sqlite3.connect(dbname)
        cursor = connection.cursor()
        table_list = cursor.execute(
            """
            SELECT name FROM sqlite_master WHERE type='table' AND name='{tale}';
            """
        ).fetchall()
        connection.commit()
        connection.close()
        return bool(table_list)



"""
Common interface for databases : MariaDB interface
"""
# pyright: reportOptionalCall=false
# pyright: reportOptionalMemberAccess=false
import importlib
import os
from datetime import datetime, timezone
from pathlib import Path
from time import sleep, time

from nua.lib.actions import python_package_installed
from nua.lib.exec import exec_as_root

from .db_manager import DbManager

if python_package_installed("mariadb"):
    mariadb = importlib.import_module("mariadb")
else:
    mariadb = None  # type:ignore

NUA_MARIADB_PWD_FILE = ".mariadb_pwd"  # noqa S105


class MariaDbManager(DbManager):
    """Database manager for MariaDB."""

    def __init__(
        self,
        host: str = "",
        port: str | int = "",
        user: str = "",
        password: str | None = None,
    ):
        if not mariadb:
            raise ValueError("The package 'mariadb' is not installed.")
        self.host = host or "localhost"
        self.port = int(port or "3306")
        self.user = user or "root"
        if password:
            self.password = password
        else:
            self.password = self.mariadb_pwd()

    def root_connect(self):
        return mariadb.connect(
            host=self.host, port=self.port, user=self.user, password=self.password
        )

    def mariadb_pwd(self) -> str:
        """Return the 'root' user DB password of mariadb.

        - container context: the env variable NUA_MARIADB_PASSWORD should contain
        the password.
        """
        pwd = os.environ.get("MARIADB_ROOT_PASSWORD")
        if not pwd:
            pwd = os.environ.get("MYSQL_ROOT_PASSWORD")
        if not pwd:
            pwd = os.environ.get("NUA_MARIADB_PASSWORD")
        if pwd:
            return pwd
        file_path = Path(os.path.expanduser("~nua")) / NUA_MARIADB_PWD_FILE
        return file_path.read_text(encoding="utf8").strip()

    def setup_db_user(self, dbname: str, user: str, password: str):
        """Create a mariadb user if needed."""
        if not self.user_exist(user):
            self.user_create(user, password)
        if not self.db_exist(dbname):
            self.db_create(dbname, user, password)

    def remove_db_user(self, dbname: str, user: str):
        """Remove a mariadb user and DB if needed."""
        if self.db_exist(dbname):
            self.db_drop(dbname)
        if self.user_exist(user):
            self.user_drop(user)

    def db_drop(self, dbname: str):
        """Basic drop database."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = f"drop database if exists {dbname}"
        cursor.execute(query)
        connection.close()

    def db_dump(self, dbname: str, **kwargs: str):
        """Basic mariadb_dump call.

        FIXME: not ok for remote host
        """
        options = kwargs.get("options", "")
        cmd = f"mariadb-dump {dbname} {options}"
        exec_as_root(cmd)

    def wait_for_db(self, timeout: int = 120):
        """Wait for the DB beeing up."""
        when = time()
        limit = when + timeout
        while time() < limit:
            while time() < when:
                sleep(0.1)
            try:
                self.root_connect()
                now = datetime.now(timezone.utc).isoformat(" ")
                print(f"{now} - Connection ok")
                return
            except mariadb.Error:
                now = datetime.now(timezone.utc).isoformat(" ")
                print(f"{now} - Connection failed")
                when += 5.0
        raise RuntimeError(f"DB not available after {timeout} seconds.")

    def user_drop(self, user: str) -> bool:
        """Drop user (wip, not enough)."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = "DROP USER IF EXISTS ?"
        cursor.execute(query, (user,))
        connection.close()
        return True

    def user_exist(self, user: str) -> bool:
        """Test if a mariadb user exists."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = "SELECT EXISTS(SELECT 1 FROM mysql.user WHERE user = ?)"  # noqa S608
        cursor.execute(query, (user,))
        (count,) = cursor.fetchone()
        connection.close()
        return count != 0

    def user_create(self, user: str, password: str):
        """Create a mariadb user."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = "CREATE USER IF NOT EXISTS ? IDENTIFIED BY ?"
        cursor.execute(query, (user, password))
        connection.commit()
        connection.close()

    def db_exist(self, dbname: str) -> bool:
        """Test if the named mariadb database exists."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = f"SHOW DATABASES LIKE '{dbname}'"
        cursor.execute(query, (dbname,))
        result = cursor.fetchone()
        connection.close()
        return bool(result)

    def db_create(self, dbname: str, user: str, password: str | None = None):
        """Create a mariadb DB."""
        connection = self.root_connect()
        cursor = connection.cursor()
        query = f"CREATE DATABASE {dbname}"
        cursor.execute(query)
        # query = f"GRANT ALL PRIVILEGES ON {dbname}.* TO '{user}'@'%' IDENTIFIED BY ?"
        query = f"GRANT ALL PRIVILEGES ON {dbname}.* TO ? IDENTIFIED BY ?"
        cursor.execute(query, (user, password))
        connection.commit()
        connection.close()

    def db_table_exist(self, dbname: str, user: str, password: str, table: str) -> bool:
        """Check if the named database exists (for host, connecting as user),
        assuming DB exists."""
        connection = mariadb.connect(
            host=self.host, database=dbname, user=user, password=password
        )
        cursor = connection.cursor()
        query = (
            "SELECT count(*) FROM information_schema.TABLES WHERE "
            f"(TABLE_SCHEMA = '{dbname}') AND (TABLE_NAME = '{table}')"
        )
        # cursor.execute(query, (dbname, table))
        cursor.execute(query)
        result = cursor.fetchone()
        count = result[0] if result else 0
        connection.close()
        return count > 0



"""Base class for detect and install project."""
import abc


class BaseDetector(abc.ABC):
    """Base class for detect and install project."""

    message = ""

    @classmethod
    def info(cls) -> str:
        return cls.message

    @classmethod
    def detect(cls) -> bool:
        return False

    @classmethod
    def install(cls) -> None:
        pass



from .version import __version__

NUA_PYTHON_TAG = f"nua-python:{__version__}"
NUA_BUILDER_TAG = f"nua-builder:{__version__}"

NUA_CONFIG_STEM = "nua-config"
NUA_CONFIG_EXT = ("json", "toml", "yaml", "yml")
# in docker:
NUA_BUILD_PATH = "/nua/build"
NUA_APP_PATH = "/nua/app"
NUA_METADATA_PATH = "/nua/metadata"
NUA_SCRIPTS_PATH = "/nua/scripts"



"""
Common interface for databases : Postgres interface

Todo: separate part for local/remote instance
"""
# pyright: reportOptionalCall=false
# pyright: reportOptionalMemberAccess=false
import importlib
import os
import re
from datetime import datetime, timezone
from pathlib import Path
from time import sleep, time

from nua.lib.actions import python_package_installed
from nua.lib.exec import mp_exec_as_postgres

from .db_manager import DbManager

if python_package_installed("psycopg2") or python_package_installed("psycopg2-binary"):
    psycopg2 = importlib.import_module("psycopg2")
    SQL = getattr(importlib.import_module("psycopg2.sql"), "SQL")  # noqa B009
    Identifier = getattr(  # noqa B009
        importlib.import_module("psycopg2.sql"), "Identifier"
    )
else:
    psycopg2 = None  # type: ignore
    SQL = None
    Identifier = None

PG_VERSION = "14"
POSTGRES_CONF_PATH = Path(f"/etc/postgresql/{PG_VERSION}/main")
RE_5432 = re.compile(r"\s*port\s*=\s*5432\D")
RE_COMMENT = re.compile(r"\s*#")
RE_LISTEN = re.compile(r"\s*listen_addresses\s*=(.*)$")
# S105 Possible hardcoded password
NUA_PG_PWD_FILE = ".postgres_pwd"  # noqa S105


class PostgresManager(DbManager):
    """Database manager for Postgres."""

    def __init__(
        self,
        host: str = "",
        port: str | int = "",
        user: str = "",
        password: str | None = None,
    ):
        if not psycopg2:
            raise ValueError("The package 'psycopg2' is not installed.")
        self.host = host or "localhost"
        self.port = int(port or "5432")
        self.user = user or "postgres"
        if password:
            self.password = password
        else:
            self.password = self.postgres_pwd()

    def __str__(self) -> str:
        return f"{self.user}@{self.host}:{self.port}"

    def root_connect(self, connect_timeout: int = 5):
        return psycopg2.connect(
            host=self.host,
            port=self.port,
            dbname="postgres",
            user=self.user,
            password=self.password,
            connect_timeout=connect_timeout,
        )

    def postgres_pwd(self) -> str:
        """Return the 'postgres' user DB password.

        - When used in container context, the env variable POSTGRES_PASSWORD or
          NUA_POSTGRES_PASSWORD should contain the password.
        - When used in nua-orchestrator context, read the password from local file.
        """
        pwd = os.environ.get("POSTGRES_PASSWORD")
        if pwd:
            # it is the common environ variable for postgres docker image
            return pwd
        pwd = os.environ.get("NUA_POSTGRES_PASSWORD")
        if pwd:
            return pwd
        file_path = Path("~nua").expanduser() / NUA_PG_PWD_FILE
        with open(file_path, encoding="utf8") as rfile:
            pwd = rfile.read().strip()
        return pwd

    def setup_db_user(self, dbname: str, user: str, password: str):
        """Create a postgres user if needed."""
        if not self.user_exist(user):
            self.user_create(user, password)
        if not self.db_exist(dbname):
            self.db_create(dbname, user)

    def remove_db_user(self, dbname: str, user: str):
        """Remove a postgres user and DB if needed."""
        if self.db_exist(dbname):
            self.db_drop(dbname)
        if self.user_exist(user):
            self.user_drop(user)

    def db_drop(self, dbname: str):
        """Basic drop database."""
        connection = self.root_connect()
        connection.autocommit = True
        with connection:
            with connection.cursor() as cur:
                query = "REVOKE CONNECT ON DATABASE {db} FROM public"
                cur.execute(SQL(query).format(db=Identifier(dbname)))
            with connection.cursor() as cur:
                query = "DROP DATABASE {db}"
                cur.execute(SQL(query).format(db=Identifier(dbname)))
        connection.close()  # <- Needed? The with statement should close it.

    def db_dump(self, dbname: str, **kwargs: str):
        """Basic pg_dump call.

        FIXME: not ok for remote host
        """
        options = kwargs.get("options", "")
        cmd = f"pg_dump {dbname} {options}"
        mp_exec_as_postgres(cmd)

    def wait_for_db(self, timeout: int = 120):
        """Wait for the DB beeing up."""
        when = time()
        limit = when + timeout
        while time() < limit:
            while time() < when:
                sleep(0.1)
            try:
                self.root_connect(connect_timeout=5)
                now = datetime.now(timezone.utc).isoformat(" ")
                print(f"{now} - Connection ok")
                return
            except psycopg2.OperationalError:
                now = datetime.now(timezone.utc).isoformat(" ")
                print(f"{now} - Connection failed")
                when += 5.0
        raise RuntimeError(f"DB not available after {timeout} seconds.")

    def user_drop(self, user: str) -> bool:
        """Drop user (wip, not enough)."""
        connection = self.root_connect()
        connection.autocommit = True
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                query = "DROP USER IF EXISTS {username}"
                cur.execute(SQL(query).format(username=Identifier(user)))
        connection.close()  # <- Needed? The with statement should close it.
        return True

    def user_exist(self, user: str) -> bool:
        """Test if a postgres user exists."""
        connection = self.root_connect()
        connection.autocommit = True
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                query = "SELECT COUNT(*) FROM pg_catalog.pg_roles WHERE rolname = %s"
                cur.execute(SQL(query), (user,))
                result = cur.fetchone()
                count = result[0] if result else 0
        connection.close()  # <- Needed? The with statement should close it.
        return count != 0

    def user_create(self, user: str, password: str):
        """Create a postgres user."""
        connection = self.root_connect()
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                # or:  CREATE USER user WITH ENCRYPTED PASSWORD 'pwd'
                query = "CREATE ROLE {username} LOGIN PASSWORD %s"
                cur.execute(SQL(query).format(username=Identifier(user)), (password,))
                cur.execute("COMMIT")
        connection.close()

    def db_exist(self, dbname: str) -> bool:
        """Test if the named postgres database exists."""
        connection = self.root_connect()
        connection.autocommit = True
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                query = "SELECT datname FROM pg_database"
                cur.execute(SQL(query))
                results = cur.fetchall()
        connection.close()
        db_set = {x[0] for x in results if x}
        return dbname in db_set

    def db_create(self, dbname: str, user: str, _password: str | None = None):
        """Create a postgres DB.

        (password unused for postgres)
        """
        connection = self.root_connect()
        connection.autocommit = True
        cur = connection.cursor()
        query = "CREATE DATABASE {db} OWNER {user}"
        cur.execute(SQL(query).format(db=Identifier(dbname), user=Identifier(user)))
        connection.close()
        connection = self.root_connect()
        connection.autocommit = True
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                # not this: WITH GRANT OPTION;"
                query = "GRANT ALL PRIVILEGES ON DATABASE {db} TO {user}"
                cur.execute(
                    SQL(query).format(db=Identifier(dbname), user=Identifier(user))
                )
        connection.close()

    def db_table_exist(self, dbname: str, user: str, password: str, table: str) -> bool:
        """Check if the named database exists (for host, connecting as user),
        assuming DB exists."""
        connection = psycopg2.connect(
            host=self.host, port=self.port, dbname=dbname, user=user, password=password
        )
        with connection:  # noqa SIM117
            with connection.cursor() as cur:
                query = (
                    "SELECT COUNT(*) FROM information_schema.tables WHERE table_name=%s"
                )
                cur.execute(SQL(query), (table,))
                result = cur.fetchone()
                count = result[0] if result else 0
        connection.close()
        return count > 0



"""Auto install Python wheels."""
from pathlib import Path

from nua.lib.actions import pip_install

from .base_detector import BaseDetector


class PythonWheels(BaseDetector):
    message = "Python wheels"

    @classmethod
    def detect(cls) -> bool:
        root = Path(".").resolve()
        return bool(list(root.glob("*.whl")))

    @classmethod
    def install(cls) -> None:
        pip_install(["*.whl"])



"""Parse the meta-packages requirements."""
import json
from functools import cache
from importlib import resources as rso

DEPS_DIR = "nua.agent.deps"
META_PACKAGES_FILE = "meta_packages.json"


@cache
def _parse_meta_packages() -> dict:
    resource_path = rso.files(DEPS_DIR).joinpath(META_PACKAGES_FILE)
    return json.loads(resource_path.read_text(encoding="utf8"))


def meta_packages_requirements(name: str) -> list:
    all_deps = _parse_meta_packages()
    return all_deps.get(name, [])



"""Install the configured application inside the container.

- information come from a mandatory local file: "nua-config.toml|json|yaml|yml"
- origin may be a source tar.gz or a git repository, python wheel
- build locally if source is python package
"""
import logging
import os
from copy import deepcopy
from pathlib import Path
from shutil import copy2
from textwrap import dedent

from nua.lib.actions import (
    apt_remove_lists,
    install_build_packages,
    install_git_source,
    install_meta_packages,
    install_packages,
    install_pip_packages,
    install_source,
)
from nua.lib.backports import chdir
from nua.lib.exec import exec_as_nua
from nua.lib.panic import Abort, info, show, vprint
from nua.lib.shell import chmod_r, chown_r, mkdir_p, rm_fr, sh
from nua.lib.tool.state import set_verbosity, verbosity, verbosity_level

from ..auto_install import detect_and_install
from ..constants import (
    NUA_APP_PATH,
    NUA_BUILD_PATH,
    NUA_METADATA_PATH,
    NUA_SCRIPTS_PATH,
)

# most inferred meta packages will provided by plugins in the future:
from ..meta_packages import meta_packages_requirements
from ..nua_config import NuaConfig, hyphen_get

logging.basicConfig(level=logging.INFO)


class BuilderApp:
    """Class to hold config and other state information during build."""

    def __init__(self):
        if "nua_verbosity" in os.environ:
            set_verbosity(int(os.environ["nua_verbosity"]))
            with verbosity(3):
                info("verbosity:", verbosity_level())

        self.build_dir = Path(NUA_BUILD_PATH)
        if not self.build_dir.is_dir():
            raise Abort(f"Build directory does not exist: '{self.build_dir}'")

        chdir(self.build_dir)
        self.config = NuaConfig(self.build_dir)
        self.source = Path()

    def build(self):
        self.make_dirs()
        with chdir(self.config.root_dir):
            self.pre_build()
            with verbosity(1):
                info("******** Stage: build")
            with install_build_packages(self.config.build_packages):
                code_installed = self.install_project_code()
                pip_installed = install_pip_packages(self.config.pip_install)
                built = False
                if code_installed:
                    built = detect_and_install(self.source)
                if not any((pip_installed, code_installed)):
                    # no package installed through install_pip_packages and
                    # no other way. Let's assume there is a local project.
                    show("Try install from some local project")
                    built = detect_and_install(".")
                if (code_installed or pip_installed or built) and os.getuid() == 0:
                    chown_r("/nua/build", "nua")
                # if code_installed:
                # always run build script: maybe no code source,
                # but only configuration of standard .deb packages
                self.run_build_script()

        self.post_build()
        self.test_build()
        with verbosity(1):
            show("******** Build done.")

    def infer_meta_packages(self) -> list:
        """Return packages inferred from the nua-config requirements."""
        inferred = []
        for resource in self.config.resource:
            inferred.extend(meta_packages_requirements(resource.get("type", "")))
        # if self.config.git_url:
        #     inferred.extend(meta_packages_requirements("git"))
        if inferred:
            with verbosity(2):
                vprint(f"Inferred meta packages: {inferred}")
        return inferred

    def pre_build(self):
        """Process installation of packages prior to running install script."""
        with verbosity(1):
            info("******** Stage: pre-build")
        install_meta_packages(self.infer_meta_packages(), keep_lists=True)
        install_meta_packages(self.config.meta_packages, keep_lists=True)
        install_packages(self.config.packages, keep_lists=True)

    def make_dirs(self):
        self._make_dir_user_nua(NUA_APP_PATH)
        self._make_dir_user_nua(NUA_METADATA_PATH)
        self.make_custom_document_root()

    def make_custom_document_root(self):
        """If the app defines a specific document root  (i.e.
        /var/www/html)."""
        document_root = hyphen_get(self.config.build, "document-root")
        if not document_root:
            return
        path = Path(document_root)
        rm_fr(path)
        self._make_dir_user_nua(path)

    @staticmethod
    def _make_dir_user_nua(directory: str | Path):
        mkdir_p(directory)
        chown_r(directory, "nua")
        chmod_r(directory, 0o644, 0o755)

    def post_build(self):
        with verbosity(1):
            info("******** Stage: post-build")
        self.merge_files()
        if os.getuid() == 0:
            chown_r("/nua/build", "nua")
        apt_remove_lists()
        self.copy_metadata()
        self.make_start_script()

    def copy_metadata(self):
        """Dump the content of the nua-config file in /nua/metadata/nua-
        config.json."""
        self.config.dump_json(NUA_METADATA_PATH)

    def make_start_script(self):
        script_dir = Path(NUA_SCRIPTS_PATH)
        script_dir.mkdir(mode=0o755, parents=True, exist_ok=True)
        if self.config.start_command:
            with verbosity(2):
                vprint("Writing start script from 'start-command'")
            return self._write_start_script(script_dir, self.config.start_command)
        path = self.find_start_script()
        if path:
            with verbosity(2):
                vprint("Copying start script:", path)
            copy2(path, script_dir)
            return
        with verbosity(2):
            vprint("Writing debug start script")
            return self._write_start_script(script_dir, ["env"])

    def _write_start_script(self, script_dir: Path, start_cmd: list):
        data = deepcopy(self.config.metadata_rendered)
        cwd = repr(str(self.source))
        cmd = dedent(
            f"""\
        import os

        from nua.lib.exec import exec_as_nua
        from nua.agent.templates import render_config_templates

        render_config_templates({data})
        exec_as_nua({start_cmd},
                    cwd={cwd},
                    env=os.environ,)
        """
        )
        script_path = script_dir / "start.py"
        script_path.write_text(cmd)

    def find_start_script(self) -> Path | None:
        name = hyphen_get(self.config.build, "start-script")
        if not name:
            name = "start.py"
        path = self.build_dir / "nua" / name
        path = path.absolute().resolve()
        if path.is_file():
            with verbosity(3):
                vprint("found start_script path:", path)
            return path
        return None

    def find_build_script(self) -> Path | None:
        name = hyphen_get(self.config.build, "build-script")
        if not name:
            name = "build.py"
        path = self.build_dir / "nua" / name
        path = path.absolute().resolve()
        if path.is_file():
            with verbosity(3):
                info(f"found build script: {path}")
            return path
        return None

    def run_build_script(self):
        """Process the 'build.py' script if exists or the build-command.

        The script is run from the directory of the nua-config.toml
        file.
        """
        if self.config.build_command:
            return self.build_command()
        script_path = self.find_build_script()
        if not script_path:
            return
        # assuming it is a python script
        env = dict(os.environ)
        with verbosity(2):
            cmd = "python --version"
            sh(cmd, env=env)
        cmd = f"python {script_path}"
        sh(cmd, env=env, timeout=1800)

    def build_command(self):
        """Process the 'build-command' commands.

        The script is run from the source directory.
        """
        if not self.config.build_command:
            return
        with chdir(self.source):
            with verbosity(2):
                show("Execution of build-command")
            exec_as_nua(self.config.build_command)

    def install_project_code(self) -> bool:
        installed = False
        if self.config.src_url:
            self.source = install_source(
                self.config.src_url,
                "/nua/build",
                self.config.name,
                self.config.checksum,
            )
            installed = True
        elif self.config.project:
            self.source = install_source(
                self.config.project,
                "/nua/build",
                self.config.name,
            )
            installed = True
        elif self.config.git_url:
            self.source = install_git_source(
                self.config.git_url,
                self.config.git_branch,
                "/nua/build",
                self.config.name,
            )
            installed = True
        return installed

    def merge_files(self):
        """Copy content of various /nua/build/nua subfolders in /nua."""
        root = Path("/nua/build/nua")
        if not root.is_dir():
            with verbosity(3):
                print("not a directory: /nua/build/nua")
            return
        for item in root.iterdir():
            if item.name == "nua" or not item.is_dir():
                continue
            for file in item.glob("**/*"):
                if not file.is_file():
                    continue
                with verbosity(3):
                    print("merge file:", item)
                target = Path("/nua").joinpath(file.relative_to(root))
                target.parent.mkdir(parents=True, exist_ok=True)
                copy2(file, target.parent)

    def test_build(self):
        """Execute a configured shell command to check build is successful."""
        default = "test -f /nua/metadata/nua-config.*"
        command = self.config.build.get("test", default)
        if not command:
            return
        show("Execution of build test command:", command)
        sh(command, show_cmd=False)


def main() -> None:
    """Setup app in Nua container."""
    builder = BuilderApp()
    builder.build()



"""Common interface for databases managers (WIP).

This could make it more uniform to use different databases
(and easier to add additional databases).

Ex:

```python

pg_manager = PostgresManager("localhost", "5432", "postgres", "postgres")
if not pg_manager.db_exists("mydb"):
    pg_manager.db_create("my_db")
...

# Where PostgresManager is a subclass of DbManager
```

TODO: refactor the existing modules to implement the API.
"""

from typing import Protocol


class DbManager(Protocol):
    """Database manager."""

    def __init__(
        self,
        host: str = "",
        port: str | int = "",
        user: str = "",
        password: str | None = None,
    ):
        ...

    # Common methods (between PostgreSQL and MariaDB)
    def setup_db_user(self, dbname: str, user: str, password: str):
        ...

    def remove_db_user(self, dbname: str, user: str):
        ...

    def db_drop(self, dbname: str):
        ...

    def db_dump(self, dbname: str, **kwargs: str):
        ...

    def user_drop(self, user: str) -> bool:
        ...

    def user_exist(self, user: str) -> bool:
        ...

    def user_create(self, user: str, password: str):
        ...

    def db_create(self, dbname: str, user: str, password: str | None = None):
        ...

    def db_exist(self, dbname: str) -> bool:
        ...

    def db_table_exist(self, dbname: str, user: str, password: str, table: str) -> bool:
        ...



"""Class to manage the "nua-config.toml" file."""
import json
import os
from copy import deepcopy
from pathlib import Path
from typing import Any, Union

import tomli
import yaml
from nua.lib.actions import download_extract
from nua.lib.shell import chown_r

from .constants import NUA_CONFIG_EXT, NUA_CONFIG_STEM
from .nua_tag import nua_tag_string

REQUIRED_BLOCKS = ["metadata"]
REQUIRED_METADATA = ["id", "version", "title", "author", "license"]
OPTIONAL_METADATA = [
    "tagline",
    "website",
    "tags",
    "profile",
    "release",
    "changelog",
    "name",
    "image",
]
# blocks added (empty) if not present in orig file:
COMPLETE_BLOCKS = ["build", "run", "env", "docker"]


class NuaConfigError(ValueError):
    pass


def nua_config_names():
    for suffix in NUA_CONFIG_EXT:
        yield f"{NUA_CONFIG_STEM}.{suffix}"


def hyphen_get(data: dict, key: str, default: Any = None) -> Any:
    """Return value from dict for key either hyphen "-" or underscore "_" in
    it, priority to "-"."""
    if (hypkey := key.replace("_", "-")) in data:
        result = data.get(hypkey)
    elif (underkey := key.replace("-", "_")) in data:
        result = data.get(underkey)
    else:
        result = default
    return result


def nomalize_env_values(env: dict[str, Union[str, int, float, list]]) -> dict[str, str]:
    def normalize_env_leaf(value: Any) -> str:
        if isinstance(value, str):
            return value
        if isinstance(value, (int, float, list)):
            return str(value)
        raise NuaConfigError(f"ENV value has wrong type: '{value}'")

    validated: dict[str, str | dict] = {}
    for key, value in env.items():
        if isinstance(value, dict):
            validated[key] = {k: normalize_env_leaf(v) for k, v in value.items()}
        else:
            validated[key] = normalize_env_leaf(value)

    return deepcopy(validated)  # type: ignore


def force_list(content: str | list[str]) -> list[str]:
    """Return always a list, if a single string is provided, wrap it into
    list."""
    if isinstance(content, list):
        return content
    if isinstance(content, str) and content:
        return [content]
    return []


class NuaConfig:
    """Class to manage the "nua-config.toml" file.

    The config file can be any of nua-config.[toml|json|yaml|yml]
    """

    path: Path
    root_dir: Path
    nua_dir_exists: bool
    _data: dict

    def __init__(self, path: str | Path | None = None):
        if not path:
            path = "."
        self._find_root_dir(path)
        self._find_config_file()
        self._loads_config()
        self._check_required_blocks()
        self._complete_missing_blocks()
        self._fix_spelling()
        self._check_required_metadata()
        self._check_checksum_format()
        self._nomalize_env_values()

    def __repr__(self):
        return f"NuaConfig(path={self.path}, id={id(self)})"

    def _find_root_dir(self, path: Path | str) -> None:
        """Find project folder.

        Analyses structure of parameters:

        Below the expected structure of folders. 'nua' folder and subfolders are
        optional, the 'nua-config' file can be at 3 positions: 'project', 'nua'
        subfolder, or 'metadata' subfolder.

        The path parameter can be:
            - the 'nua-config' file
            - the path of 'nua' folder, 'project' folder or folder containing the
              'nua-config' file

        We try to:
            - find the 'nua-file'
            - find the 'project' path
            - find the 'nua' path (if exists)

        The 'nua-config' file used is the one of higher level found.

        /project
            ...project code source...
            nua-config
            /nua
                Dockerfile
                build.py
                ...
                nua-config
                /metadata
                    nua-config
                /build
                   ...files to copy...
                /app
                   ...files to copy...
        """
        path = Path(path).resolve()
        if path.is_dir():
            folder = path
        else:
            folder = path.parent
        if folder.name == "nua":
            self.root_dir = folder.parent
        elif folder.name == "metadata" and folder.parent.name == "nua":
            self.root_dir = folder.parent.parent
        else:
            self.root_dir = folder
        self.nua_dir_exists = (self.root_dir / "nua").is_dir()

    def _find_config_file(self):
        """Find nua-config file."""
        for folder in (
            self.root_dir,
            self.root_dir / "nua",
            self.root_dir / "nua" / "metadata",
        ):
            if not folder.is_dir():
                continue
            for name in nua_config_names():
                test_path = folder / name
                if test_path.is_file():
                    self.path = test_path
                    return

        raise NuaConfigError(
            f"Nua config file not found in '{self.root_dir}' and sub folders"
        )

    def _loads_config(self):
        if self.path.suffix == ".toml":
            self._data = tomli.loads(self.path.read_text(encoding="utf8"))
        elif self.path.suffix in {".json", ".yaml", ".yml"}:
            self._data = yaml.safe_load(self.path.read_text(encoding="utf8"))
        else:
            raise NuaConfigError(f"Unknown file extension for '{self.path}'")

    def as_dict(self) -> dict:
        return self._data

    def dump_json(self, folder: Path | str) -> None:
        dest = Path(folder) / f"{NUA_CONFIG_STEM}.json"
        dest.write_text(
            json.dumps(self._data, sort_keys=False, ensure_ascii=False, indent=4),
            encoding="utf8",
        )

    def _check_required_blocks(self):
        for block in REQUIRED_BLOCKS:
            if block not in self._data:
                raise NuaConfigError(
                    f"Missing mandatory block in {self.path}: '{block}'"
                )

    def _complete_missing_blocks(self):
        for block in COMPLETE_BLOCKS:
            if block not in self._data:
                self._data[block] = {}

    def _fix_spelling(self):
        if "license" not in self.metadata and "licence" in self.metadata:
            self.metadata["license"] = self.metadata["licence"]

    def _check_required_metadata(self):
        for key in REQUIRED_METADATA:
            if key not in self._data["metadata"]:
                raise NuaConfigError(
                    f"Missing mandatory metadata in {self.path}: '{key}'"
                )

    def _check_checksum_format(self):
        checksum = self.checksum
        if not checksum:
            return
        if checksum.startswith("sha256:"):
            checksum = checksum[7:]
        if len(checksum) != 64:
            raise NuaConfigError(
                f"Wrong checksum content (expecting 64 length sha256): {checksum}"
            )
        self.metadata["checksum"] = checksum

    def _nomalize_env_values(self):
        self._data["env"] = nomalize_env_values(self.env)

    def __getitem__(self, key: str) -> Any:
        """will return {} is key not found, assuming some parts are not
        mandatory and first level element are usually dict."""
        return self._data.get(key) or {}

    # metadata ######################################################

    @property
    def metadata(self) -> dict:
        return self["metadata"]

    @property
    def metadata_rendered(self) -> dict:
        """Return the metadata dict with rendered f-string values."""
        data = {}
        for key, val in deepcopy(self.metadata).items():
            if isinstance(val, str):
                data[key] = val.format(**self.metadata)
            else:
                data[key] = val
        return data

    @property
    def version(self) -> str:
        """Version of package source."""
        return self.metadata.get("version", "")

    @property
    def src_url(self) -> str:
        if base := hyphen_get(self.metadata, "src-url"):
            return base.format(**self.metadata)
        return ""

    @property
    def git_url(self) -> str:
        if base := hyphen_get(self.metadata, "git-url"):
            return base.format(**self.metadata)
        return ""

    @property
    def git_branch(self) -> str:
        if base := hyphen_get(self.metadata, "git-branch"):
            return base.format(**self.metadata)
        return "main"

    @property
    def wrap_image(self) -> str:
        """Optional  Docker 'image' to be used as base for 'wrap' strategy.

        If the 'image' metadata is defined, the build strategy is to add
        the '/nua/metadata/nua-config.json' file on the declared image,
        when build method is 'wrap'.
        """
        if base := self.metadata.get("image", ""):
            return base.format(**self.metadata)
        return ""

    @property
    def checksum(self) -> str:
        """Return checksum associated to 'src-url' or null string."""
        return self.metadata.get("checksum", "").strip().lower()

    @property
    def app_id(self) -> str:
        return self.metadata.get("id", "")

    @property
    def name(self) -> str:
        if name := self.metadata.get("name", ""):
            return name
        return self.app_id

    @property
    def nua_tag(self) -> str:
        return nua_tag_string(self.metadata)

    # build #########################################################

    @property
    def build(self) -> dict:
        return self["build"]

    @property
    def builder(self) -> str:
        return self.build.get("builder", "")

    @property
    def project(self) -> str:
        """The project URL to build with autodetection."""
        if base := self.build.get("project"):
            return base.format(**self.metadata)
        return ""

    @property
    def manifest(self) -> list:
        return force_list(self.build.get("manifest", []))

    @property
    def meta_packages(self) -> list:
        return force_list(hyphen_get(self.build, "meta-packages", []))

    @property
    def build_packages(self) -> list:
        return force_list(self.build.get("packages", []))

    @property
    def build_command(self) -> list:
        """Return the list of build commands, each cmd rendered with
        metadata."""
        metadata = self.metadata_rendered
        commands = []
        for cmd in force_list(hyphen_get(self.build, "build-command", [])):
            if isinstance(cmd, str):
                commands.append(cmd.format(**metadata))
            else:
                commands.append(cmd)
        return commands

    @property
    def pip_install(self) -> list:
        return force_list(hyphen_get(self.build, "pip-install", []))

    @property
    def build_method(self) -> str:
        """Build method (or default build method).

        Can be empty string if not defined (then autodetection from
        metadata).
        """
        default_method = hyphen_get(self.build, "default-method", "")
        return self.build.get("method", default_method)

    @property
    def docker_user(self) -> str:
        """User of the Docker container, default is root.

        Especially usefull when wrapping an existing Dockerfile.
        """
        return hyphen_get(self.build, "docker-user", "")

    # run ###########################################################

    @property
    def run(self) -> dict:
        return self["run"]

    @property
    def packages(self) -> list:
        "Return list of run packages (packages remanent after build)."
        return force_list(self.run.get("packages", []))

    @property
    def start_command(self) -> list:
        return force_list(hyphen_get(self.run, "start", []))

    # env ###########################################################

    @property
    def env(self) -> dict:
        return self["env"]

    # docker env ####################################################

    @property
    def docker(self) -> dict:
        return self["docker"]

    # resource ######################################################

    @property
    def resource(self) -> list:
        """The list of resources (tag 'resource')."""
        return self._data.get("resource", [])

    # actions #######################################################

    def fetch_source(self, name: str = "") -> Path:
        """Download, check and extract source URL, returning the source Path.

        Set ownership to 'nua' user if launched by 'root'.
        """
        if not name:
            name = self.name
            if name.startswith("nua-"):
                name = name[4:]
        path = download_extract(self.src_url, "/nua/build", name, self.checksum)
        if os.getuid() == 0:
            chown_r(path, "nua")
        return path



"""Generate the long nua tag of packages.

For Docker container and network names, the ":" will be replaced by "-".
"""


def nua_tag_string(metadata: dict) -> str:
    """Return long tag string with version and release.

    "hedgedoc" -> "nua-hedgedoc:1.9.7-3"
    """
    if release := metadata.get("release", ""):
        suffix = f"-{release}"
    else:
        suffix = ""
    if metadata["id"].startswith("nua-"):
        prefix = ""
    else:
        prefix = "nua-"
    return f"{prefix}{metadata['id']}:{metadata['version']}{suffix}"
